@@ -1,6 +1,8 @@
 /* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */
 
 const Clutter = imports.gi.Clutter;
+const GLib = imports.gi.GLib;
+const Gio = imports.gi.Gio;
 const Gtk = imports.gi.Gtk;
 const Lang = imports.lang;
 const Mainloop = imports.mainloop;
@@ -14,6 +16,7 @@ const Tweener = imports.ui.tweener;
 const Main = imports.ui.main;
 const BoxPointer = imports.ui.boxpointer;
 const Params = imports.misc.params;
+const Utils = imports.misc.utils;
 
 const ANIMATION_TIME = 0.2;
 const NOTIFICATION_TIMEOUT = 4;
@@ -22,10 +25,12 @@ const SUMMARY_TIMEOUT = 1;
 const HIDE_TIMEOUT = 0.2;
 const LONGER_HIDE_TIMEOUT = 0.6;
 
-const BUTTON_ICON_SIZE = 36;
-
 const MAX_SOURCE_TITLE_WIDTH = 180;
 
+// We delay hiding of the tray if the mouse is within MOUSE_LEFT_ACTOR_THRESHOLD
+// range from the point where it left the tray.
+const MOUSE_LEFT_ACTOR_THRESHOLD = 20;
+
 const State = {
     HIDDEN:  0,
     SHOWING: 1,
@@ -33,15 +38,132 @@ const State = {
     HIDING:  3
 };
 
-function _cleanMarkup(text) {
-    // Support &amp;, &quot;, &apos;, &lt; and &gt;, escape all other
-    // occurrences of '&'.
-    let _text = text.replace(/&(?!amp;|quot;|apos;|lt;|gt;)/g, '&amp;');
-    // Support <b>, <i>, and <u>, escape anything else
-    // so it displays as raw markup.
-    return _text.replace(/<(\/?[^biu]>|[^>\/][^>])/g, '&lt;$1');
+function _fixMarkup(text, allowMarkup) {
+    if (allowMarkup) {
+        // Support &amp;, &quot;, &apos;, &lt; and &gt;, escape all other
+        // occurrences of '&'.
+        let _text = text.replace(/&(?!amp;|quot;|apos;|lt;|gt;)/g, '&amp;');
+        // Support <b>, <i>, and <u>, escape anything else
+        // so it displays as raw markup.
+        return _text.replace(/<(\/?[^biu]>|[^>\/][^>])/g, '&lt;$1');
+    } else {
+        // Escape everything
+        let _text = text.replace(/&/g, '&amp;');
+        return _text.replace(/</g, '&lt;');
+    }
 }
 
+function URLHighlighter(text, lineWrap, allowMarkup) {
+    this._init(text, lineWrap, allowMarkup);
+}
+
+URLHighlighter.prototype = {
+    _init: function(text, lineWrap, allowMarkup) {
+        if (!text)
+            text = '';
+        this.actor = new St.Label({ reactive: true, style_class: 'url-highlighter' });
+        this._linkColor = '#ccccff';
+        this.actor.connect('style-changed', Lang.bind(this, function() {
+            let color = new Clutter.Color();
+            let hasColor = this.actor.get_theme_node().get_color('link-color', color);
+            if (hasColor) {
+                let linkColor = color.to_string().substr(0, 7);
+                if (linkColor != this._linkColor) {
+                    this._linkColor = linkColor;
+                    this._highlightUrls();
+                }
+            }
+        }));
+        if (lineWrap) {
+            this.actor.clutter_text.line_wrap = true;
+            this.actor.clutter_text.line_wrap_mode = Pango.WrapMode.WORD_CHAR;
+            this.actor.clutter_text.ellipsize = Pango.EllipsizeMode.NONE;
+        }
+
+        this.setMarkup(text, allowMarkup);
+        this.actor.connect('button-release-event', Lang.bind(this, function (actor, event) {
+            let urlId = this._findUrlAtPos(event);
+            if (urlId != -1) {
+                let url = this._urls[urlId].url;
+                if (url.indexOf(':') == -1)
+                    url = 'http://' + url;
+                try {
+                    Gio.app_info_launch_default_for_uri(url, global.create_app_launch_context());
+                    return true;
+                } catch (e) {
+                    // TODO: remove this after gnome 3 release
+                    let p = new Shell.Process({ 'args' : ['gvfs-open', url] });
+                    p.run();
+                    return true;
+                }
+            }
+            return false;
+        }));
+        this.actor.connect('motion-event', Lang.bind(this, function(actor, event) {
+            let urlId = this._findUrlAtPos(event);
+            if (urlId != -1 && !this._cursorChanged) {
+                global.set_cursor(Shell.Cursor.POINTING_HAND);
+                this._cursorChanged = true;
+            } else if (urlId == -1) {
+                global.unset_cursor();
+                this._cursorChanged = false;
+            }
+            return false;
+        }));
+        this.actor.connect('leave-event', Lang.bind(this, function() {
+            if (this._cursorChanged) {
+                this._cursorChanged = false;
+                global.unset_cursor();
+            }
+        }));
+    },
+
+    setMarkup: function(text, allowMarkup) {
+        text = text ? _fixMarkup(text, allowMarkup) : '';
+        this._text = text;
+
+        this.actor.clutter_text.set_markup(text);
+        /* clutter_text.text contain text without markup */
+        this._urls = Utils.findUrls(this.actor.clutter_text.text);
+        this._highlightUrls();
+    },
+
+    _highlightUrls: function() {
+        // text here contain markup
+        let urls = Utils.findUrls(this._text);
+        let markup = '';
+        let pos = 0;
+        for (let i = 0; i < urls.length; i++) {
+            let url = urls[i];
+            let str = this._text.substr(pos, url.pos - pos);
+            markup += str + '<span foreground="' + this._linkColor + '"><u>' + url.url + '</u></span>';
+            pos = url.pos + url.url.length;
+        }
+        markup += this._text.substr(pos);
+        this.actor.clutter_text.set_markup(markup);
+    },
+
+    _findUrlAtPos: function(event) {
+        let success;
+        let [x, y] = event.get_coords();
+        [success, x, y] = this.actor.transform_stage_point(x, y);
+        let find_pos = -1;
+        for (let i = 0; i < this.actor.clutter_text.text.length; i++) {
+            let [success, px, py, line_height] = this.actor.clutter_text.position_to_coords(i);
+            if (py > y || py + line_height < y || x < px)
+                continue;
+            find_pos = i;
+        }
+        if (find_pos != -1) {
+            for (let i = 0; i < this._urls.length; i++)
+            if (find_pos >= this._urls[i].pos &&
+                this._urls[i].pos + this._urls[i].url.length > find_pos)
+                return i;
+        }
+        return -1;
+    }
+};
+
 // Notification:
 // @source: the notification's Source
 // @title: the title
@@ -71,7 +193,8 @@ function _cleanMarkup(text) {
 // area.
 //
 // @params can contain values for 'customContent', 'body', 'icon',
-// and 'clear' parameters.
+// 'titleMarkup', 'bannerMarkup', 'bodyMarkup', and 'clear'
+// parameters.
 //
 // If @params contains a 'customContent' parameter with the value %true,
 // then @banner will not be shown in the body of the notification when the
@@ -85,6 +208,12 @@ function _cleanMarkup(text) {
 // source.createNotificationIcon(). However, if @params contains an 'icon'
 // parameter, the passed in icon will be used.
 //
+// If @params contains a 'titleMarkup', 'bannerMarkup', or
+// 'bodyMarkup' parameter with the value %true, then the corresponding
+// element is assumed to use pango markup. If the parameter is not
+// present for an element, then anything that looks like markup in
+// that element will appear literally in the output.
+//
 // If @params contains a 'clear' parameter with the value %true, then
 // the content and the action area of the notification will be cleared.
 // The content area is also always cleared if 'customContent' is false
@@ -97,9 +226,14 @@ Notification.prototype = {
     _init: function(source, title, banner, params) {
         this.source = source;
         this.urgent = false;
+        this.resident = false;
+        // 'transient' is a reserved keyword in JS, so we have to use an alternate variable name
+        this.isTransient = false;
         this.expanded = false;
+        this._useActionIcons = false;
         this._customContent = false;
         this._bannerBodyText = null;
+        this._bannerBodyMarkup = false;
         this._titleFitsInBannerMode = true;
         this._spacing = 0;
 
@@ -111,7 +245,6 @@ Notification.prototype = {
         this._prevFocusedWindow = null;
         this._prevKeyFocusActor = null;
 
-        this._focusWindowChangedId = 0;
         this._focusActorChangedId = 0;
         this._stageInputModeChangedId = 0;
         this._capturedEventId = 0;
@@ -126,7 +259,7 @@ Notification.prototype = {
             function (actor, event) {
                 if (!this._actionArea ||
                     !this._actionArea.contains(event.get_source()))
-                    this.emit('clicked');
+                    this._onClicked();
             }));
 
         // The first line should have the title, followed by the
@@ -145,7 +278,8 @@ Notification.prototype = {
 
         this._titleLabel = new St.Label();
         this._bannerBox.add_actor(this._titleLabel);
-        this._bannerLabel = new St.Label();
+        this._bannerUrlHighlighter = new URLHighlighter();
+        this._bannerLabel = this._bannerUrlHighlighter.actor;
         this._bannerBox.add_actor(this._bannerLabel);
 
         this.update(title, banner, params);
@@ -172,6 +306,9 @@ Notification.prototype = {
         params = Params.parse(params, { customContent: false,
                                         body: null,
                                         icon: null,
+                                        titleMarkup: false,
+                                        bannerMarkup: false,
+                                        bodyMarkup: false,
                                         clear: false });
 
         this._customContent = params.customContent;
@@ -202,7 +339,7 @@ Notification.prototype = {
                                      y_fill: false,
                                      y_align: St.Align.START });
 
-        title = title ? _cleanMarkup(title.replace(/\n/g, ' ')) : '';
+        title = title ? _fixMarkup(title.replace(/\n/g, ' '), params.titleMarkup) : '';
         this._titleLabel.clutter_text.set_markup('<b>' + title + '</b>');
 
         // Unless the notification has custom content, we save this._bannerBodyText
@@ -210,9 +347,11 @@ Notification.prototype = {
         // expandable due to other elements in its content area or due to the banner
         // not fitting fully in the single-line mode.
         this._bannerBodyText = this._customContent ? null : banner;
+        this._bannerBodyMarkup = params.bannerMarkup;
 
-        banner = banner ? _cleanMarkup(banner.replace(/\n/g, '  ')) : '';
-        this._bannerLabel.clutter_text.set_markup(banner);
+        banner = banner ? banner.replace(/\n/g, '  ') : '';
+
+        this._bannerUrlHighlighter.setMarkup(banner, params.bannerMarkup);
         this._bannerLabel.queue_relayout();
 
         // Add the bannerBody now if we know for sure we'll need it
@@ -220,7 +359,7 @@ Notification.prototype = {
             this._addBannerBody();
 
         if (params.body)
-            this.addBody(params.body);
+            this.addBody(params.body, params.bodyMarkup);
         this._updated();
     },
 
@@ -228,7 +367,7 @@ Notification.prototype = {
     // @actor: actor to add to the body of the notification
     //
     // Appends @actor to the notification's body
-    addActor: function(actor) {
+    addActor: function(actor, style) {
         if (!this._scrollArea) {
             this.actor.add_style_class_name('multi-line-notification');
             this._scrollArea = new St.ScrollView({ name: 'notification-scrollview',
@@ -245,34 +384,30 @@ Notification.prototype = {
             this._addBannerBody();
         }
 
-        this._contentArea.add(actor);
+        this._contentArea.add(actor, style ? style : {});
         this._updated();
     },
 
     // addBody:
     // @text: the text
+    // @markup: %true if @text contains pango markup
+    // @style: style to use when adding the actor containing the text
     //
     // Adds a multi-line label containing @text to the notification.
     //
     // Return value: the newly-added label
-    addBody: function(text) {
-        let body = new St.Label();
-        body.clutter_text.line_wrap = true;
-        body.clutter_text.line_wrap_mode = Pango.WrapMode.WORD_CHAR;
-        body.clutter_text.ellipsize = Pango.EllipsizeMode.NONE;
+    addBody: function(text, markup, style) {
+        let label = new URLHighlighter(text, true, markup);
 
-        text = text ? _cleanMarkup(text) : '';
-        body.clutter_text.set_markup(text);
-
-        this.addActor(body);
-        return body;
+        this.addActor(label.actor, style);
+        return label.actor;
     },
 
     _addBannerBody: function() {
         if (this._bannerBodyText) {
             let text = this._bannerBodyText;
             this._bannerBodyText = null;
-            this.addBody(text);
+            this.addBody(text, this._bannerBodyMarkup);
         }
     },
 
@@ -342,16 +477,16 @@ Notification.prototype = {
 
         let button = new St.Button();
 
-        if (Gtk.IconTheme.get_default().has_icon(id)) {
+        if (this._useActionIcons && Gtk.IconTheme.get_default().has_icon(id)) {
             button.add_style_class_name('notification-icon-button');
-            button.child = St.TextureCache.get_default().load_icon_name(id, St.IconType.SYMBOLIC, BUTTON_ICON_SIZE);
+            button.child = new St.Icon({ icon_name: id });
         } else {
             button.add_style_class_name('notification-button');
             button.label = label;
         }
 
         this._buttonBox.add(button);
-        button.connect('clicked', Lang.bind(this, function() { this.emit('action-invoked', id); }));
+        button.connect('clicked', Lang.bind(this, this._onActionInvoked, id));
         this._updated();
     },
 
@@ -359,6 +494,18 @@ Notification.prototype = {
         this.urgent = urgent;
     },
 
+    setResident: function(resident) {
+        this.resident = resident;
+    },
+
+    setTransient: function(isTransient) {
+        this.isTransient = isTransient;
+    },
+
+    setUseActionIcons: function(useIcons) {
+        this._useActionIcons = useIcons;
+    },
+
     _styleChanged: function() {
         this._spacing = this.actor.get_theme_node().get_length('spacing-columns');
     },
@@ -476,29 +623,16 @@ Notification.prototype = {
         let metaDisplay = global.screen.get_display();
 
         this._prevFocusedWindow = metaDisplay.focus_window;
-        this._prevKeyFocus = global.stage.get_key_focus();
+        this._prevKeyFocusActor = global.stage.get_key_focus();
 
-        // We need to use the captured event in the overview, because we don't want to change the stage input mode to
-        // FOCUSED there. On the other hand, using the captured event doesn't work correctly in the main view because
-        // it doesn't allow focusing the windows again correctly. So we are using the FOCUSED stage input mode in the
-        // main view.
-        if (Main.overview.visible) {
-            if (!Main.pushModal(this.actor))
-                return;
-            this._capturedEventId = global.stage.connect('captured-event', Lang.bind(this, this._onCapturedEvent));
-        } else {
+        if (!Main.overview.visible)
             global.set_stage_input_mode(Shell.StageInputMode.FOCUSED);
 
-            this._focusWindowChangedId = metaDisplay.connect('notify::focus-window', Lang.bind(this, this._focusWindowChanged));
-            this._stageInputModeChangedId = global.connect('notify::stage-input-mode', Lang.bind(this, this._stageInputModeChanged));
+        // Use captured-event to notice clicks outside the notification
+        // without consuming them.
+        this._capturedEventId = global.stage.connect('captured-event', Lang.bind(this, this._onCapturedEvent));
 
-            this._keyPressId = global.stage.connect('key-press-event', Lang.bind(this, this._onKeyPress));
-        }
-
-        // We need to listen to this signal in the overview, as well as in the main view, to make the key bindings such as
-        // Alt+F2 work. When a notification has key focus, which is the case with chat notifications, all captured KEY_PRESS
-        // events have the actor with the key focus as their source. This makes it impossible to distinguish between the chat
-        // window input and the key bindings based solely on the KEY_PRESS event.
+        this._stageInputModeChangedId = global.connect('notify::stage-input-mode', Lang.bind(this, this._stageInputModeChanged));
         this._focusActorChangedId = global.stage.connect('notify::key-focus', Lang.bind(this, this._focusActorChanged));
 
         this._hasFocus = true;
@@ -506,19 +640,6 @@ Notification.prototype = {
             Main.messageTray.lock();
     },
 
-    _focusWindowChanged: function() {
-        let metaDisplay = global.screen.get_display();
-        // this._focusWindowChanged() will be called when we call
-        // global.set_stage_input_mode(Shell.StageInputMode.FOCUSED) ,
-        // however metaDisplay.focus_window will be null in that case. We only
-        // want to ungrab focus if the focus has been moved to an application
-        // window.
-        if (metaDisplay.focus_window) {
-            this._prevFocusedWindow = null;
-            this.ungrabFocus();
-        }
-    },
-
     _focusActorChanged: function() {
         let focusedActor = global.stage.get_key_focus();
         if (!focusedActor || !this.actor.contains(focusedActor)) {
@@ -528,15 +649,6 @@ Notification.prototype = {
     },
 
     _stageInputModeChanged: function() {
-        let focusedActor = global.stage.get_key_focus();
-        // TODO: We need to set this._prevFocusedWindow to null in order to
-        // get the cursor in the run dialog. However, that also means it's
-        // set to null when the application menu is activated, which defeats
-        // the point of keeping the name of the previously focused application
-        // in the panel. It'd be good to be able to distinguish between these
-        // two cases.
-        this._prevFocusedWindow = null;
-        this._prevKeyFocusActor = null;
         this.ungrabFocus();
     },
 
@@ -559,6 +671,28 @@ Notification.prototype = {
         return false;
     },
 
+    _onActionInvoked: function(actor, id) {
+        this.emit('action-invoked', id);
+        if (!this.resident) {
+            // We don't hide a resident notification when the user invokes one of its actions,
+            // because it is common for such notifications to update themselves with new
+            // information based on the action. We'd like to display the updated information
+            // in place, rather than pop-up a new notification.
+            this.emit('done-displaying');
+            this.destroy();
+        }
+    },
+
+    _onClicked: function() {
+        this.emit('clicked');
+        // We hide all types of notifications once the user clicks on them because the common
+        // outcome of clicking should be the relevant window being brought forward and the user's
+        // attention switching to the window.
+        this.emit('done-displaying');
+        if (!this.resident)
+            this.destroy();
+    },
+
     _onKeyPress: function(actor, event) {
         let symbol = event.get_key_symbol();
         if (symbol == Clutter.Escape) {
@@ -573,10 +707,6 @@ Notification.prototype = {
             return;
 
         let metaDisplay = global.screen.get_display();
-        if (this._focusWindowChangedId > 0) {
-            metaDisplay.disconnect(this._focusWindowChangedId);
-            this._focusWindowChangedId = 0;
-        }
 
         if (this._focusActorChangedId > 0) {
             global.stage.disconnect(this._focusActorChangedId);
@@ -589,20 +719,14 @@ Notification.prototype = {
         }
 
         if (this._capturedEventId > 0) {
-            Main.popModal(this.actor);
             global.stage.disconnect(this._capturedEventId);
             this._capturedEventId = 0;
         }
 
-        if (this._keyPressId > 0) {
-            global.stage.disconnect(this._keyPressId);
-            this._keyPressId = 0;
-        }
-
         this._hasFocus = false;
         Main.messageTray.unlock();
 
-        if (this._prevFocusedWindow) {
+        if (this._prevFocusedWindow && !metaDisplay.focus_window) {
             metaDisplay.set_input_focus_window(this._prevFocusedWindow, false, global.get_current_time());
             this._prevFocusedWindow = null;
         }
@@ -644,6 +768,11 @@ Source.prototype = {
         this.title = title;
         this._iconBin = new St.Bin({ width: this.ICON_SIZE,
                                      height: this.ICON_SIZE });
+        this.isTransient = false;
+    },
+
+    setTransient: function(isTransient) {
+        this.isTransient = isTransient;
     },
 
     // Called to create a new icon actor (of size this.ICON_SIZE).
@@ -674,6 +803,7 @@ Source.prototype = {
                     this.notification = null;
                     this._notificationDestroyedId = 0;
                     this._notificationClickedId = 0;
+                    this._notificationRemoved();
                 }
             }));
 
@@ -693,8 +823,13 @@ Source.prototype = {
         this._iconBin.child = icon;
     },
 
-    // Default implementation is to do nothing, but subclass can override
+    // Default implementation is to do nothing, but subclasses can override
     _notificationClicked: function(notification) {
+    },
+
+    // Default implementation is to destroy this source, but subclasses can override
+    _notificationRemoved: function() {
+        this.destroy();
     }
 };
 Signals.addSignalMethods(Source.prototype);
@@ -768,6 +903,7 @@ MessageTray.prototype = {
         this._notificationBin.hide();
         this._notificationQueue = [];
         this._notification = null;
+        this._notificationClickedId = 0;
 
         this._summaryBin = new St.Bin({ anchor_gravity: Clutter.Gravity.NORTH_EAST });
         this.actor.add_actor(this._summaryBin);
@@ -789,6 +925,7 @@ MessageTray.prototype = {
         this._summaryNotificationBoxPointer.actor.hide();
 
         this._summaryNotification = null;
+        this._summaryNotificationClickedId = 0;
         this._clickedSummaryItem = null;
         this._clickedSummaryItemAllocationChangedId = 0;
         this._expandedSummaryItem = null;
@@ -896,7 +1033,15 @@ MessageTray.prototype = {
         }
 
         this._summaryItems.push(summaryItem);
-        this._newSummaryItems.push(summaryItem);
+
+        // We keep this._newSummaryItems to track any new sources that were added to the
+        // summary and show the summary with them to the user for a short period of time
+        // after notifications are done showing. However, we don't want that to happen for
+        // transient sources, which are removed after the notification is shown, but are
+        // not removed fast enough because of the callbacks to avoid the summary popping up.
+        // So we just don't add transient sources to this._newSummaryItems .
+        if (!source.isTransient)
+            this._newSummaryItems.push(summaryItem);
 
         source.connect('notify', Lang.bind(this, this._onNotify));
 
@@ -1146,6 +1291,8 @@ MessageTray.prototype = {
             if (this._trayLeftTimeoutId) {
                 Mainloop.source_remove(this._trayLeftTimeoutId);
                 this._trayLeftTimeoutId = 0;
+                this._trayLeftMouseX = -1;
+                this._trayLeftMouseY = -1;
                 return;
             }
 
@@ -1167,6 +1314,14 @@ MessageTray.prototype = {
             this._pointerInTray = true;
             this._updateState();
         } else {
+            // We record the position of the mouse the moment it leaves the tray. These coordinates are used in
+            // this._onTrayLeftTimeout() to determine if the mouse has moved far enough during the initial timeout for us
+            // to consider that the user intended to leave the tray and therefore hide the tray. If the mouse is still
+            // close to its previous position, we extend the timeout once.
+            let [x, y, mods] = global.get_pointer();
+            this._trayLeftMouseX = x;
+            this._trayLeftMouseY = y;
+
             // We wait just a little before hiding the message tray in case the user quickly moves the mouse back into it.
             // We wait for a longer period if the notification popped up where the mouse pointer was already positioned.
             // That gives the user more time to mouse away from the notification and mouse back in in order to expand it.
@@ -1176,12 +1331,25 @@ MessageTray.prototype = {
     },
 
     _onTrayLeftTimeout: function() {
-        this._useLongerTrayLeftTimeout = false;
-        this._trayLeftTimeoutId = 0;
-        this._pointerInTray = false;
-        this._pointerInSummary = false;
-        this._updateNotificationTimeout(0);
-        this._updateState();
+        let [x, y, mods] = global.get_pointer();
+        // We extend the timeout once if the mouse moved no further than MOUSE_LEFT_ACTOR_THRESHOLD to either side or up.
+        // We don't check how far down the mouse moved because any point above the tray, but below the exit coordinate,
+        // is close to the tray.
+        if (this._trayLeftMouseX > -1 &&
+            y > this._trayLeftMouseY - MOUSE_LEFT_ACTOR_THRESHOLD &&
+            x < this._trayLeftMouseX + MOUSE_LEFT_ACTOR_THRESHOLD &&
+            x > this._trayLeftMouseX - MOUSE_LEFT_ACTOR_THRESHOLD) {
+            this._trayLeftMouseX = -1;
+            this._trayLeftTimeoutId = Mainloop.timeout_add(LONGER_HIDE_TIMEOUT * 1000,
+                                                             Lang.bind(this, this._onTrayLeftTimeout));
+        } else {
+            this._trayLeftTimeoutId = 0;
+            this._useLongerTrayLeftTimeout = false;
+            this._pointerInTray = false;
+            this._pointerInSummary = false;
+            this._updateNotificationTimeout(0);
+            this._updateState();
+        }
         return false;
     },
 
@@ -1305,6 +1473,8 @@ MessageTray.prototype = {
 
     _showNotification: function() {
         this._notification = this._notificationQueue.shift();
+        this._notificationClickedId = this._notification.connect('done-displaying',
+                                                                 Lang.bind(this, this.escapeTray));
         this._notificationBin.child = this._notification.actor;
 
         this._notificationBin.opacity = 0;
@@ -1402,7 +1572,12 @@ MessageTray.prototype = {
         this._notificationBin.hide();
         this._notificationBin.child = null;
         this._notification.collapseCompleted();
+        this._notification.disconnect(this._notificationClickedId);
+        this._notificationClickedId = 0;
+        let notification = this._notification;
         this._notification = null;
+        if (notification.isTransient)
+            notification.destroy();
     },
 
     _expandNotification: function(autoExpanding) {
@@ -1476,7 +1651,8 @@ MessageTray.prototype = {
 
     _showSummaryNotification: function() {
         this._summaryNotification = this._clickedSummaryItem.source.notification;
-
+        this._summaryNotificationClickedId = this._summaryNotification.connect('done-displaying',
+                                                                               Lang.bind(this, this.escapeTray));
         let index = this._notificationQueue.indexOf(this._summaryNotification);
         if (index != -1)
             this._notificationQueue.splice(index, 1);
@@ -1499,8 +1675,8 @@ MessageTray.prototype = {
         this._summaryNotificationBoxPointer.actor.show();
         this._adjustNotificationBoxPointerPosition();
 
-        this._summaryNotificationState = State.SHOWNING;
-        this._summaryNotificationBoxPointer.animateAppear(Lang.bind(this, function() {
+        this._summaryNotificationState = State.SHOWING;
+        this._summaryNotificationBoxPointer.show(true, Lang.bind(this, function() {
             this._summaryNotificationState = State.SHOWN;
         }));
     },
@@ -1539,15 +1715,19 @@ MessageTray.prototype = {
 
         this._summaryNotification.ungrabFocus();
         this._summaryNotificationState = State.HIDING;
-        this._summaryNotificationBoxPointer.animateDisappear(Lang.bind(this, this._hideSummaryNotificationCompleted));
+        this._summaryNotificationBoxPointer.hide(true, Lang.bind(this, this._hideSummaryNotificationCompleted));
     },
 
     _hideSummaryNotificationCompleted: function() {
         this._summaryNotificationState = State.HIDDEN;
         this._summaryNotificationBoxPointer.bin.child = null;
         this._summaryNotification.collapseCompleted();
+        this._summaryNotification.disconnect(this._summaryNotificationClickedId);
+        this._summaryNotificationClickedId = 0;
         let summaryNotification = this._summaryNotification;
         this._summaryNotification = null;
+        if (summaryNotification.isTransient && !this._reNotifyWithSummaryNotificationAfterHide)
+            summaryNotification.destroy();
         if (this._reNotifyWithSummaryNotificationAfterHide) {
             this._onNotify(summaryNotification.source, summaryNotification);
             this._reNotifyWithSummaryNotificationAfterHide = false;
