@@ -1,26 +1,22 @@
 /* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */
 
-const Cairo = imports.cairo;
 const Clutter = imports.gi.Clutter;
-const GConf = imports.gi.GConf;
 const Gio = imports.gi.Gio;
-const Gtk = imports.gi.Gtk;
 const Lang = imports.lang;
 const Mainloop = imports.mainloop;
-const Meta = imports.gi.Meta;
 const Pango = imports.gi.Pango;
 const Shell = imports.gi.Shell;
 const St = imports.gi.St;
 const Signals = imports.signals;
-const DBus = imports.dbus;
 const Gettext = imports.gettext.domain('gnome-shell');
 const _ = Gettext.gettext;
 
-const AppDisplay = imports.ui.appDisplay;
-const BoxPointer = imports.ui.boxpointer;
 const Calendar = imports.ui.calendar;
+const Config = imports.misc.config;
 const Overview = imports.ui.overview;
 const PopupMenu = imports.ui.popupMenu;
+const PanelMenu = imports.ui.panelMenu;
+const StatusMenu = imports.ui.statusMenu;
 const Main = imports.ui.main;
 const Tweener = imports.ui.tweener;
 
@@ -34,14 +30,16 @@ const ANIMATED_ICON_UPDATE_TIMEOUT = 100;
 const SPINNER_UPDATE_TIMEOUT = 130;
 const SPINNER_SPEED = 0.02;
 
-const STANDARD_TRAY_ICON_ORDER = ['keyboard', 'volume', 'bluetooth', 'network', 'battery'];
-const STANDARD_TRAY_ICON_IMPLEMENTATIONS = {
-    'bluetooth-applet': 'bluetooth',
-    'gnome-volume-control-applet': 'volume',
-    'nm-applet': 'network',
-    'gnome-power-manager': 'battery'
+const STANDARD_TRAY_ICON_ORDER = ['a11y', 'display', 'keyboard', 'volume', 'bluetooth', 'network', 'battery'];
+const STANDARD_TRAY_ICON_SHELL_IMPLEMENTATION = {
+    'a11y': imports.ui.status.accessibility.ATIndicator,
+    'volume': imports.ui.status.volume.Indicator,
+    'battery': imports.ui.status.power.Indicator
 };
 
+if (Config.HAVE_BLUETOOTH)
+    STANDARD_TRAY_ICON_SHELL_IMPLEMENTATION['bluetooth'] = imports.ui.status.bluetooth.Indicator;
+
 const CLOCK_FORMAT_KEY        = 'format';
 const CLOCK_CUSTOM_FORMAT_KEY = 'custom-format';
 const CLOCK_SHOW_DATE_KEY     = 'show-date';
@@ -176,38 +174,6 @@ TextShadower.prototype = {
     }
 };
 
-function PanelMenuButton(menuAlignment) {
-    this._init(menuAlignment);
-}
-
-PanelMenuButton.prototype = {
-    _init: function(menuAlignment) {
-        this.actor = new St.Bin({ style_class: 'panel-button',
-                                  reactive: true,
-                                  x_fill: true,
-                                  y_fill: false,
-                                  track_hover: true });
-        this.actor._delegate = this;
-        this.actor.connect('button-press-event', Lang.bind(this, this._onButtonPress));
-        this.menu = new PopupMenu.PopupMenu(this.actor, menuAlignment, St.Side.TOP, /* FIXME */ 0);
-        this.menu.connect('open-state-changed', Lang.bind(this, this._onOpenStateChanged));
-        Main.chrome.addActor(this.menu.actor, { visibleInOverview: true,
-                                                affectsStruts: false });
-        this.menu.actor.hide();
-    },
-
-    _onButtonPress: function(actor, event) {
-        this.menu.toggle();
-    },
-
-    _onOpenStateChanged: function(menu, open) {
-        if (open)
-            this.actor.add_style_pseudo_class('pressed');
-        else
-            this.actor.remove_style_pseudo_class('pressed');
-    }
-};
-
 /**
  * AppMenuButton:
  *
@@ -221,13 +187,14 @@ function AppMenuButton() {
 }
 
 AppMenuButton.prototype = {
-    __proto__: PanelMenuButton.prototype,
+    __proto__: PanelMenu.Button.prototype,
 
     _init: function() {
-        PanelMenuButton.prototype._init.call(this, St.Align.START);
+        PanelMenu.Button.prototype._init.call(this, St.Align.START);
         this._metaDisplay = global.screen.get_display();
+        this._startingApps = [];
 
-        this._focusedApp = null;
+        this._targetApp = null;
 
         let bin = new St.Bin({ name: 'appMenu' });
         this.actor.set_child(bin);
@@ -258,10 +225,11 @@ AppMenuButton.prototype = {
 
         this._updateId = 0;
         this._animationStep = 0;
-        this._clipWidth = AppDisplay.APPICON_SIZE / 2;
+        this._clipWidth = PANEL_ICON_SIZE;
         this._direction = SPINNER_SPEED;
 
-        this._spinner = new AnimatedIcon('process-working.png', 24);
+        this._spinner = new AnimatedIcon('process-working.png',
+                                         PANEL_ICON_SIZE);
         this._container.add_actor(this._spinner.actor);
         this._spinner.actor.lower_bottom();
 
@@ -273,6 +241,8 @@ AppMenuButton.prototype = {
         tracker.connect('notify::focus-app', Lang.bind(this, this._sync));
         tracker.connect('app-state-changed', Lang.bind(this, this._onAppStateChanged));
 
+        global.window_manager.connect('switch-workspace', Lang.bind(this, this._sync));
+
         this._sync();
     },
 
@@ -343,7 +313,7 @@ AppMenuButton.prototype = {
         }
         if (this._animationStep > 1)
             this._animationStep = 1;
-        this._clipWidth = this._label.actor.width - (this._label.actor.width - AppDisplay.APPICON_SIZE / 2) * (1 - this._animationStep);
+        this._clipWidth = this._label.actor.width - (this._label.actor.width - PANEL_ICON_SIZE) * (1 - this._animationStep);
         if (this.actor.get_direction() == St.TextDirection.LTR) {
             this._label.actor.set_clip(0, 0, this._clipWidth + this._shadow.width, this.actor.height);
         } else {
@@ -443,18 +413,19 @@ AppMenuButton.prototype = {
     },
 
     _onQuit: function() {
-        if (this._focusedApp == null)
+        if (this._targetApp == null)
             return;
-        this._focusedApp.request_quit();
+        this._targetApp.request_quit();
     },
 
     _onAppStateChanged: function(tracker, app) {
         let state = app.state;
-        if (app == this._lastStartedApp
-            && state != Shell.AppState.STARTING) {
-            this._lastStartedApp = null;
+        if (state != Shell.AppState.STARTING) {
+            this._startingApps = this._startingApps.filter(function(a) {
+                return a != app;
+            });
         } else if (state == Shell.AppState.STARTING) {
-            this._lastStartedApp = app;
+            this._startingApps.push(app);
         }
         // For now just resync on all running state changes; this is mainly to handle
         // cases where the focused window's application changes without the focus
@@ -465,14 +436,27 @@ AppMenuButton.prototype = {
 
     _sync: function() {
         let tracker = Shell.WindowTracker.get_default();
+        let lastStartedApp = null;
+        let workspace = global.screen.get_active_workspace();
+        for (let i = 0; i < this._startingApps.length; i++)
+            if (this._startingApps[i].is_on_workspace(workspace))
+                lastStartedApp = this._startingApps[i];
 
         let focusedApp = tracker.focus_app;
-        if (focusedApp == this._focusedApp) {
-            if (focusedApp && focusedApp.get_state() != Shell.AppState.STARTING)
+        let targetApp = focusedApp != null ? focusedApp : lastStartedApp;
+        if (targetApp == this._targetApp) {
+            if (targetApp && targetApp.get_state() != Shell.AppState.STARTING)
                 this.stopAnimation();
             return;
-        } else {
-            this._stopAnimation();
+        }
+        this._stopAnimation();
+
+        if (!focusedApp) {
+            // If the app has just lost focus to the panel, pretend
+            // nothing happened; otherwise you can't keynav to the
+            // app menu.
+            if (global.stage_input_mode == Shell.StageInputMode.FOCUSED)
+                return;
         }
 
         if (this._iconBox.child != null)
@@ -481,15 +465,13 @@ AppMenuButton.prototype = {
         this._label.setText('');
         this.actor.reactive = false;
 
-        this._focusedApp = focusedApp;
-
-        let targetApp = this._focusedApp != null ? this._focusedApp : this._lastStartedApp;
+        this._targetApp = targetApp;
         if (targetApp != null) {
-            let icon = targetApp.get_faded_icon(AppDisplay.APPICON_SIZE);
+            let icon = targetApp.get_faded_icon(2 * PANEL_ICON_SIZE);
 
             this._label.setText(targetApp.get_name());
             // TODO - _quit() doesn't really work on apps in state STARTING yet
-            this._quitMenu.label.set_text(_('Quit %s').format(targetApp.get_name()));
+            this._quitMenu.label.set_text(_("Quit %s").format(targetApp.get_name()));
 
             this.actor.reactive = true;
             this._iconBox.set_child(icon);
@@ -510,10 +492,10 @@ function ClockButton() {
 }
 
 ClockButton.prototype = {
-    __proto__: PanelMenuButton.prototype,
+    __proto__: PanelMenu.Button.prototype,
 
     _init: function() {
-        PanelMenuButton.prototype._init.call(this, St.Align.START);
+        PanelMenu.Button.prototype._init.call(this, St.Align.START);
         this.menu.addAction(_("Preferences"), Lang.bind(this, this._onPrefs));
 
         this._clock = new St.Label();
@@ -530,11 +512,14 @@ ClockButton.prototype = {
 
     _onButtonPress: function(actor, event) {
         let button = event.get_button();
-        if (button == 3 &&
-            (!this._calendarPopup || !this._calendarPopup.isOpen))
+        let menuShowing = this.menu.isOpen;
+        let calendarShowing = this._calendarPopup && this._calendarPopup.isOpen;
+
+        if (menuShowing || (button == 3 && !calendarShowing))
             this.menu.toggle();
         else
             this._toggleCalendar();
+        return true;
     },
 
     closeCalendar: function() {
@@ -543,15 +528,12 @@ ClockButton.prototype = {
 
         this._calendarPopup.hide();
 
-        this.menu.isOpen = false;
         this.actor.remove_style_pseudo_class('pressed');
     },
 
     openCalendar: function() {
         this._calendarPopup.show();
 
-        // simulate an open menu, so it won't appear beneath the calendar
-        this.menu.isOpen = true;
         this.actor.add_style_pseudo_class('pressed');
     },
 
@@ -657,6 +639,13 @@ Panel.prototype = {
                                         reactive: true });
         this.actor._delegate = this;
 
+        Main.overview.connect('shown', Lang.bind(this, function () {
+            this.actor.add_style_class_name('in-overview');
+        }));
+        Main.overview.connect('hiding', Lang.bind(this, function () {
+            this.actor.remove_style_class_name('in-overview');
+        }));
+
         this._menus = new PopupMenu.PopupMenuManager(this);
 
         this._leftBox = new St.BoxLayout({ name: 'panelLeft' });
@@ -759,8 +748,9 @@ Panel.prototype = {
         /* Translators: If there is no suitable word for "Activities" in your language, you can use the word for "Overview". */
         let label = new St.Label({ text: _("Activities") });
         this.button = new St.Clickable({ name: 'panelActivities',
-                                          style_class: 'panel-button',
-                                          reactive: true });
+                                         style_class: 'panel-button',
+                                         reactive: true,
+                                         can_focus: true });
         this.button.set_child(label);
 
         this._leftBox.add(this.button);
@@ -813,46 +803,25 @@ Panel.prototype = {
         this._clockButton = new ClockButton();
         this._centerBox.add(this._clockButton.actor, { y_fill: true });
 
-        this._menus.addMenu(this._clockButton.menu);
+        let clockMenuManager = new PopupMenu.PopupMenuManager(this);
+        clockMenuManager.addMenu(this._clockButton.menu);
 
         /* right */
 
-        // Yet-another-Ubuntu-workaround - we have to kill their
-        // app-indicators, so that applications fall back to normal
-        // status icons
-        // http://bugzilla.gnome.org/show_bug.cgi=id=621382
-        let p = new Shell.Process({ args: ['pkill', '-f',
-                                           '^([^ ]*/)?indicator-application-service$']});
-        p.run();
-
-        // The tray icons live in trayBox within trayContainer.
+        // System status applets live in statusBox, while legacy tray icons
+        // live in trayBox
         // The trayBox is hidden when there are no tray icons.
-        let trayContainer = new St.Bin({ y_align: St.Align.MIDDLE });
-        this._rightBox.add(trayContainer);
-        let trayBox = new St.BoxLayout({ name: 'statusTray' });
+        let statusBox = new St.BoxLayout({ name: 'statusTray' });
+        let trayBox = new St.BoxLayout({ name: 'legacyTray' });
         this._trayBox = trayBox;
+        this._statusBox = statusBox;
 
         trayBox.hide();
-        trayContainer.add_actor(trayBox);
+        this._rightBox.add(trayBox);
+        this._rightBox.add(statusBox);
 
-        this._traymanager = new Shell.TrayManager();
-        this._traymanager.connect('tray-icon-added', Lang.bind(this, this._onTrayIconAdded));
-        this._traymanager.connect('tray-icon-removed',
-            Lang.bind(this, function(o, icon) {
-                trayBox.remove_actor(icon);
-
-                if (trayBox.get_children().length == 0)
-                    trayBox.hide();
-                this._recomputeTraySize();
-            }));
-        this._traymanager.manage_stage(global.stage);
-
-        // We need to do this here to avoid a circular import with
-        // prototype dependencies.
-        let StatusMenu = imports.ui.statusMenu;
-        this._statusmenu = new StatusMenu.StatusMenuButton();
-        this._menus.addMenu(this._statusmenu.menu);
-        this._rightBox.add(this._statusmenu.actor);
+        Main.statusIconDispatcher.connect('status-icon-added', Lang.bind(this, this._onTrayIconAdded));
+        Main.statusIconDispatcher.connect('status-icon-removed', Lang.bind(this, this._onTrayIconRemoved));
 
         // TODO: decide what to do with the rest of the panel in the Overview mode (make it fade-out, become non-reactive, etc.)
         // We get into the Overview mode on button-press-event as opposed to button-release-event because eventually we'll probably
@@ -879,6 +848,24 @@ Panel.prototype = {
         Main.chrome.addActor(this.actor, { visibleInOverview: true });
     },
 
+    startStatusArea: function() {
+        for (let i = 0; i < STANDARD_TRAY_ICON_ORDER.length; i++) {
+            let role = STANDARD_TRAY_ICON_ORDER[i];
+            let constructor = STANDARD_TRAY_ICON_SHELL_IMPLEMENTATION[role];
+            if (!constructor) {
+                // This icon is not implemented (this is a bug)
+                continue;
+            }
+            let indicator = new constructor();
+            this._statusBox.add(indicator.actor);
+            this._menus.addMenu(indicator.menu);
+        }
+
+        this._statusmenu = new StatusMenu.StatusMenuButton();
+        this._menus.addMenu(this._statusmenu.menu);
+        this._rightBox.add(this._statusmenu.actor);
+    },
+
     hideCalendar: function() {
         this._clockButton.closeCalendar();
     },
@@ -892,49 +879,40 @@ Panel.prototype = {
                          });
     },
 
-    _onTrayIconAdded: function(o, icon, wmClass) {
+    _onTrayIconAdded: function(o, icon, role) {
         icon.height = PANEL_ICON_SIZE;
 
-        let role = STANDARD_TRAY_ICON_IMPLEMENTATIONS[wmClass];
-        if (!role) {
-            // Unknown icons go first in undefined order
+        if (STANDARD_TRAY_ICON_SHELL_IMPLEMENTATION[role]) {
+            // This icon is legacy, and replaced by a Shell version
+            // Hide it
+            return;
+        }
+        // Figure out the index in our well-known order for this icon
+        let position = STANDARD_TRAY_ICON_ORDER.indexOf(role);
+        icon._rolePosition = position;
+        let children = this._trayBox.get_children();
+        let i;
+        // Walk children backwards, until we find one that isn't
+        // well-known, or one where we should follow
+        for (i = children.length - 1; i >= 0; i--) {
+            let rolePosition = children[i]._rolePosition;
+            if (!rolePosition || position > rolePosition) {
+                this._trayBox.insert_actor(icon, i + 1);
+                break;
+            }
+        }
+        if (i == -1) {
+            // If we didn't find a position, we must be first
             this._trayBox.insert_actor(icon, 0);
-        } else {
-            icon._role = role;
-            // Figure out the index in our well-known order for this icon
-            let position = STANDARD_TRAY_ICON_ORDER.indexOf(role);
-            icon._rolePosition = position;
-            let children = this._trayBox.get_children();
-            let i;
-            // Walk children backwards, until we find one that isn't
-            // well-known, or one where we should follow
-            for (i = children.length - 1; i >= 0; i--) {
-                let rolePosition = children[i]._rolePosition;
-                if (!rolePosition || position > rolePosition) {
-                    this._trayBox.insert_actor(icon, i + 1);
-                    break;
-                }
-            }
-            if (i == -1) {
-                // If we didn't find a position, we must be first
-                this._trayBox.insert_actor(icon, 0);
-            }
         }
 
         // Make sure the trayBox is shown.
         this._trayBox.show();
-        this._recomputeTraySize();
     },
 
-    // By default, tray icons have a spacing of TRAY_SPACING.  However this
-    // starts to fail if we have too many as can sadly happen; just jump down
-    // to a spacing of 8 if we're over 6.
-    // http://bugzilla.gnome.org/show_bug.cgi?id=590495
-    _recomputeTraySize: function () {
-        if (this._trayBox.get_children().length > 6)
-            this._trayBox.add_style_pseudo_class('compact');
-        else
-            this._trayBox.remove_style_pseudo_class('compact');
+    _onTrayIconRemoved: function(o, icon) {
+        if (icon.get_parent() != null)
+            this._trayBox.remove_actor(icon);
     },
 
     _addRipple : function(delay, time, startScale, startOpacity, finalScale, finalOpacity) {
