@@ -3,6 +3,7 @@
 const DBus = imports.dbus;
 const GLib = imports.gi.GLib;
 const Lang = imports.lang;
+const Mainloop = imports.mainloop;
 const Signals = imports.signals;
 const St = imports.gi.St;
 const Gettext = imports.gettext.domain('gnome-shell');
@@ -16,6 +17,7 @@ let contactManager;
 let channelDispatcher;
 
 // See Notification.appendMessage
+const SCROLLBACK_IMMEDIATE_TIME = 60; // 1 minute
 const SCROLLBACK_RECENT_TIME = 15 * 60; // 15 minutes
 const SCROLLBACK_RECENT_LENGTH = 20;
 const SCROLLBACK_IDLE_LENGTH = 5;
@@ -42,6 +44,10 @@ subscribedContactsChannel[Telepathy.CHANNEL_NAME + '.ChannelType'] = Telepathy.C
 subscribedContactsChannel[Telepathy.CHANNEL_NAME + '.TargetHandleType'] = Telepathy.HandleType.LIST;
 subscribedContactsChannel[Telepathy.CHANNEL_NAME + '.TargetID'] = 'subscribe';
 
+const NotificationDirection = {
+    SENT: 'chat-sent',
+    RECEIVED: 'chat-received'
+};
 
 // This is GNOME Shell's implementation of the Telepathy 'Client'
 // interface. Specifically, the shell is a Telepathy 'Observer', which
@@ -347,7 +353,9 @@ ContactManager.prototype = {
             let uri = GLib.filename_to_uri(file, null);
             iconBox.child = textureCache.load_uri_async(uri, iconBox._size, iconBox._size);
         } else {
-            iconBox.child = textureCache.load_icon_name('stock_person', St.IconType.FULLCOLOR, iconBox._size);
+            iconBox.child = new St.Icon({ icon_name: 'stock_person',
+                                          icon_type: St.IconType.FULLCOLOR,
+                                          icon_size: iconBox._size });
         }
     },
 
@@ -467,11 +475,14 @@ Source.prototype = {
                 }));
         }
 
+        this._notification = new Notification(this);
+
         // Since we only create sources when receiving a message, this
         // is a plausible default
         this._presence = Telepathy.ConnectionPresenceType.AVAILABLE;
 
         this._channelText = new Telepathy.ChannelText(DBus.session, connName, channelPath);
+        this._sentId = this._channelText.connect('Sent', Lang.bind(this, this._messageSent));
         this._receivedId = this._channelText.connect('Received', Lang.bind(this, this._messageReceived));
 
         this._channelText.ListPendingMessagesRemote(false, Lang.bind(this, this._gotPendingMessages));
@@ -515,22 +526,27 @@ Source.prototype = {
     _channelClosed: function() {
         this._channel.disconnect(this._closedId);
         this._channelText.disconnect(this._receivedId);
+        this._channelText.disconnect(this._sentId);
         this.destroy();
     },
 
-    _ensureNotification: function() {
-        if (!Main.messageTray.contains(this))
-            Main.messageTray.add(this);
-
-        if (!this._notification)
-            this._notification = new Notification(this);
-    },
-
     _messageReceived: function(channel, id, timestamp, sender,
                                type, flags, text) {
-        this._ensureNotification();
-        this._notification.appendMessage(text);
-        this.notify(this._notification);
+        this._notification.appendMessage(text, timestamp, NotificationDirection.RECEIVED);
+        this.notify();
+    },
+
+    // This is called for both messages we send from
+    // our client and other clients as well.
+    _messageSent: function(channel, timestamp, type, text) {
+        this._notification.appendMessage(text, timestamp, NotificationDirection.SENT);
+    },
+
+    notify: function() {
+        if (!Main.messageTray.contains(this))
+            Main.messageTray.add(this);
+
+        MessageTray.Source.prototype.notify.call(this, this._notification);
     },
 
     respond: function(text) {
@@ -538,22 +554,22 @@ Source.prototype = {
     },
 
     setPresence: function(presence, message) {
-        let msg, notify;
+        let msg, shouldNotify;
 
         if (presence == Telepathy.ConnectionPresenceType.AVAILABLE) {
             msg = _("%s is online.").format(this.title);
-            notify = (this._presence == Telepathy.ConnectionPresenceType.OFFLINE);
+            shouldNotify = (this._presence == Telepathy.ConnectionPresenceType.OFFLINE);
         } else if (presence == Telepathy.ConnectionPresenceType.OFFLINE ||
                    presence == Telepathy.ConnectionPresenceType.EXTENDED_AWAY) {
             presence = Telepathy.ConnectionPresenceType.OFFLINE;
             msg = _("%s is offline.").format(this.title);
-            notify = (this._presence != Telepathy.ConnectionPresenceType.OFFLINE);
+            shouldNotify = (this._presence != Telepathy.ConnectionPresenceType.OFFLINE);
         } else if (presence == Telepathy.ConnectionPresenceType.AWAY) {
             msg = _("%s is away.").format(this.title);
-            notify = false;
+            shouldNotify = false;
         } else if (presence == Telepathy.ConnectionPresenceType.BUSY) {
             msg = _("%s is busy.").format(this.title);
-            notify = false;
+            shouldNotify = false;
         } else
             return;
 
@@ -562,10 +578,9 @@ Source.prototype = {
         if (message)
             msg += ' <i>(' + GLib.markup_escape_text(message, -1) + ')</i>';
 
-        this._ensureNotification();
-        this._notification.appendMessage(msg, true);
-        if (notify)
-            this.notify(this._notification);
+        this._notification.appendPresence(msg, shouldNotify);
+        if (shouldNotify)
+            this.notify();
     }
 };
 
@@ -578,29 +593,47 @@ Notification.prototype = {
 
     _init: function(source) {
         MessageTray.Notification.prototype._init.call(this, source, source.title, null, { customContent: true });
+        this.setResident(true);
 
         this._responseEntry = new St.Entry({ style_class: 'chat-response' });
         this._responseEntry.clutter_text.connect('activate', Lang.bind(this, this._onEntryActivated));
         this.setActionArea(this._responseEntry);
 
         this._history = [];
+        this._timestampTimeoutId = 0;
     },
 
-    appendMessage: function(text, asTitle) {
-        if (asTitle)
-            this.update(text, null, { customContent: true });
-        else
-            this.update(this.source.title, text, { customContent: true });
-        this._append(text, 'chat-received');
+    appendMessage: function(text, timestamp, direction) {
+        this.update(this.source.title, text, { customContent: true });
+        this._append(text, direction, timestamp);
     },
 
-    _append: function(text, style) {
+    _append: function(text, style, timestamp) {
+        let currentTime = (Date.now() / 1000);
+        if (!timestamp)
+            timestamp = currentTime;
+        let lastMessageTime = -1;
+        if (this._history.length > 0)
+            lastMessageTime = this._history[0].time;
+
+        // Reset the old message timeout
+        if (this._timestampTimeoutId)
+            Mainloop.source_remove(this._timestampTimeoutId);
+
         let body = this.addBody(text);
         body.add_style_class_name(style);
         this.scrollTo(St.Side.BOTTOM);
 
-        let now = new Date().getTime() / 1000;
-        this._history.unshift({ actor: body, time: now });
+        this._history.unshift({ actor: body, time: timestamp, realMessage: true });
+
+        if (timestamp < currentTime - SCROLLBACK_IMMEDIATE_TIME)
+            this._appendTimestamp();
+        else
+            // Schedule a new timestamp in SCROLLBACK_IMMEDIATE_TIME
+            // from the timestamp of the message.
+            this._timestampTimeoutId = Mainloop.timeout_add_seconds(
+                SCROLLBACK_IMMEDIATE_TIME - (currentTime - timestamp),
+                Lang.bind(this, this._appendTimestamp));
 
         if (this._history.length > 1) {
             // Keep the scrollback from growing too long. If the most
@@ -609,17 +642,44 @@ Notification.prototype = {
             // SCROLLBACK_RECENT_LENGTH previous messages. Otherwise
             // we'll keep SCROLLBACK_IDLE_LENGTH messages.
 
-            let lastMessageTime = this._history[1].time;
-            let maxLength = (lastMessageTime < now - SCROLLBACK_RECENT_TIME) ?
+            let maxLength = (lastMessageTime < currentTime - SCROLLBACK_RECENT_TIME) ?
                 SCROLLBACK_IDLE_LENGTH : SCROLLBACK_RECENT_LENGTH;
-            if (this._history.length > maxLength) {
-                let expired = this._history.splice(maxLength);
+            let filteredHistory = this._history.filter(function(item) { return item.realMessage });
+            if (filteredHistory.length > maxLength) {
+                let lastMessageToKeep = filteredHistory[maxLength];
+                let expired = this._history.splice(this._history.indexOf(lastMessageToKeep));
                 for (let i = 0; i < expired.length; i++)
                     expired[i].actor.destroy();
             }
         }
     },
 
+    _appendTimestamp: function() {
+        let lastMessageTime = this._history[0].time;
+        let lastMessageDate = new Date(lastMessageTime * 1000);
+
+        /* Translators: this is a time format string followed by a date.
+           If applicable, replace %X with a strftime format valid for your
+           locale, without seconds. */
+        // xgettext:no-c-format
+        let timeLabel = this.addBody(lastMessageDate.toLocaleFormat(_("Sent at %X on %A")), false, { expand: true, x_fill: false, x_align: St.Align.END });
+        timeLabel.add_style_class_name('chat-meta-message');
+        this._history.unshift({ actor: timeLabel, time: lastMessageTime, realMessage: false });
+
+        this._timestampTimeoutId = 0;
+        return false;
+    },
+
+    appendPresence: function(text, asTitle) {
+        if (asTitle)
+            this.update(text, null, { customContent: true });
+        else
+            this.update(this.source.title, null, { customContent: true });
+        let label = this.addBody(text);
+        label.add_style_class_name('chat-meta-message');
+        this._history.unshift({ actor: label, time: (Date.now() / 1000), realMessage: false});
+    },
+
     grabFocus: function(lockTray) {
         // Need to call the base class function first so that
         // it saves where the key focus was before.
@@ -632,8 +692,9 @@ Notification.prototype = {
         if (text == '')
             return;
 
+        // Telepathy sends out the Sent signal for us.
+        // see Source._messageSent
         this._responseEntry.set_text('');
-        this._append(text, 'chat-sent');
         this.source.respond(text);
     }
 };
